@@ -24,148 +24,243 @@ import urllib.parse
 import urllib.request
 
 from SPARQLWrapper import __version__ as sparqlwrapper_version
+from urllib.parse import urlparse, parse_qs, urlencode
+from requests.adapters import HTTPAdapter, Retry
 
 from esmecata.utils import get_rest_uniprot_release, get_sparql_uniprot_release, is_valid_dir, send_uniprot_sparql_query, urllib_query
 from esmecata import __version__ as esmecata_version
 
 URLLIB_HEADERS = {'User-Agent': 'EsMeCaTa annotation v' + esmecata_version + ', request by urllib package v' + urllib.request.__version__}
 
+POLLING_INTERVAL = 3
+API_URL = "https://rest.uniprot.org"
+
 logger = logging.getLogger(__name__)
 
 
-def rest_query_uniprot_to_retrieve_function(protein_queries, beta=None):
+# Set of Python functions from https://www.uniprot.org/help/id_mapping
+
+def check_response(response):
+    """ Check reponse status, if error raise it.
+    Function from: https://www.uniprot.org/help/id_mapping
+
+    Args:
+        response (requests Reponse object): response returns by request query.
+    """
+    try:
+        response.raise_for_status()
+    except requests.HTTPError:
+        print(response.json())
+        raise
+
+
+def get_id_mapping_results_link(session, job_id):
+    """ Get URL containing results from mapping request to UniProt.
+    Function from: https://www.uniprot.org/help/id_mapping
+
+    Args:
+        session (requests Session object): session used to query UniProt.
+        job_id (str): Job ID corresponding to the request ID given by UniProt.
+
+    Returns:
+        str: redirect URL to the results of the mapping query.
+    """
+    url = f"{API_URL}/idmapping/details/{job_id}"
+    request = session.get(url)
+    check_response(request)
+    return request.json()["redirectURL"]
+
+
+def get_next_link(headers):
+    """ From batch queries, get the next link to request.
+    Function from: https://www.uniprot.org/help/id_mapping
+
+    Args:
+        headers (dict): headers of batch response.
+
+    Returns:
+        str: next link to query in the batch process.
+    """
+    re_next_link = re.compile(r'<(.+)>; rel="next"')
+    if "Link" in headers:
+        match = re_next_link.match(headers["Link"])
+        if match:
+            return match.group(1)
+
+
+def get_batch(session, batch_response):
+    """ Batch queries.
+    Function from: https://www.uniprot.org/help/id_mapping
+
+    Args:
+        session (requests Session object): session used to query UniProt.
+        batch_response (requests Response object): response to batch query.
+
+    Returns:
+        json: batch reponses in json.
+    """
+    batch_url = get_next_link(batch_response.headers)
+    while batch_url:
+        batch_response = session.get(batch_url)
+        batch_response.raise_for_status()
+        yield batch_response.json()
+        batch_url = get_next_link(batch_response.headers)
+
+
+def combine_batches(all_results, batch_results):
+    """ Combine the response of all batch queries.
+    Function from: https://www.uniprot.org/help/id_mapping
+
+    Args:
+        all_results (json): results from first query.
+        batch_results (json): batch results.
+
+    Returns:
+        all_results (json): combined batch reponses in json.
+    """
+    for key in ("results", "failedIds"):
+        if key in batch_results and batch_results[key]:
+            all_results[key] += batch_results[key]
+
+    return all_results
+
+
+def print_progress_batches(batch_index, size, total):
+    """ print the progress of the download.
+    Function from: https://www.uniprot.org/help/id_mapping
+
+    Args:
+        batch_index (str): index of batch process in all batch results.
+        size (int): size of the results.
+        total (int):  total size of the batch results.
+    """
+    n_fetched = min((batch_index + 1) * size, total)
+    logger.info(f"|EsMeCaTa|annotation| Fetched: {n_fetched} / {total}")
+
+
+def get_id_mapping_results_search(session, url):
+    """ Retrieve UniProt mapping results.
+    Function from: https://www.uniprot.org/help/id_mapping
+
+    Args:
+        session (requests Session object): session used to query UniProt.
+        url (str): Uniprot mapping URL containing the results.
+
+    Returns:
+        results (json): mapping results.
+    """
+    parsed = urlparse(url)
+    query = parse_qs(parsed.query)
+
+    if "size" in query:
+        size = int(query["size"][0])
+    else:
+        size = 500
+        query["size"] = size
+
+    parsed = parsed._replace(query=urlencode(query, doseq=True))
+    url = parsed.geturl()
+    request = session.get(url)
+    check_response(request)
+    results = request.json()
+    #total = int(request.headers["x-total-results"])
+    #print_progress_batches(0, size, total)
+    for i, batch in enumerate(get_batch(session, request), 1):
+        results = combine_batches(results, batch)
+        #print_progress_batches(i, size, total)
+
+    return results
+
+
+def rest_query_uniprot_to_retrieve_function(protein_queries):
     """REST query to get annotation from proteins.
 
     Args:
         protein_queries (list): list of proteins
-        beta (bool): option to use the new API of UniProt (in beta can be unstable)
 
     Returns:
         output_dict (dict): annotation dict: protein as key and annotation as value ([function_name, review_status, [go_terms], [ec_numbers], [interpros], [rhea_ids], gene_name])
     """
     output_dict = {}
 
+    url = 'http://rest.uniprot.org/idmapping/run'
+
     # Column names can be found at: https://www.uniprot.org/help/uniprotkb_column_names
     # from and to are linked to mapping ID: https://www.uniprot.org/help/api%5Fidmapping
-    if not beta:
-        url = 'https://www.uniprot.org/uploadlists/'
+    retries = Retry(total=5, backoff_factor=0.25, status_forcelist=[500, 502, 503, 504])
+    session = requests.Session()
+    session.mount("https://", HTTPAdapter(max_retries=retries))
+    params = {
+        'from': 'UniProtKB_AC-ID',
+        'to': 'UniProtKB',
+        'ids': protein_queries
+    }
+    data = urllib.parse.urlencode(params)
+    data = data.encode('utf-8')
+    req = urllib.request.Request(url, data)
 
-        params = {
-            'from': 'ACC+ID',
-            'to': 'ACC',
-            'format': 'tab',
-            'query': protein_queries,
-            'columns': 'id,protein names,reviewed,go,ec,interpro(db_abbrev),rhea-id,genes(PREFERRED)'
-        }
+    data_js = json.load(urllib_query(req))
+    job_id = data_js['jobId']
 
-        go_pattern = re.compile(r'\[GO:(?P<go>\d{7})\]')
-        data = urllib.parse.urlencode(params)
-        data = data.encode('utf-8')
-        req = urllib.request.Request(url, data, headers=URLLIB_HEADERS)
+    http_check_status = 'http://rest.uniprot.org/idmapping/status/{0}'.format(job_id)
 
-        with urllib_query(req) as http_response:
-            response_text = http_response.read().decode('utf-8')
-            if response_text != '':
-                csvreader = csv.reader(response_text.splitlines(), delimiter='\t')
-                # Avoid header
-                next(csvreader)
-            else:
-                csvreader = []
+    check_status_response = requests.head(http_check_status)
+    check_status_response.raise_for_status()
+
+    # Not working because the code is 200 instead of 303...
+    # Issue with protein_queries formatting?
+    # Wait 5 secodns to check again return code.
+    # But is there no better way?
+    while check_status_response.status_code != 303:
+        time.sleep(5)
+        check_status_response = requests.head(http_check_status)
+
+    if check_status_response.status_code == 303:
+        link = get_id_mapping_results_link(session, job_id)
+        data = get_id_mapping_results_search(session, link)
 
         results = {}
-        for line in csvreader:
-            protein_name = line[1]
-            if line[2] == 'reviewed':
+        for result in data['results']:
+            protein_id = result['from']
+            protein_data = result['to']
+            reviewed = protein_data['entryType']
+            if 'reviewed' in reviewed:
                 review = True
             else:
                 review = False
-            go_result = go_pattern.findall(line[3])
-            gos = ['GO:'+ go_term for go_term in go_result]
-            ec_numbers = [ec for ec in line[4].split('; ') if ec != '']
-            interpros =  [interpro for interpro in line[5].split('; ') if interpro != '']
-            rhea_ids = [rhea_id for rhea_id in line[6].split('; ') if rhea_id != '']
-            gene_name = line[7]
-            results[line[0]] = [protein_name, review, gos, ec_numbers, interpros, rhea_ids, gene_name]
-            output_dict.update(results)
-    else:
-        url = 'http://rest.uniprot.org/idmapping/run'
-
-        # Column names can be found at: https://www.uniprot.org/help/uniprotkb_column_names
-        # from and to are linked to mapping ID: https://www.uniprot.org/help/api%5Fidmapping
-
-        params = {
-            'from': 'UniProtKB_AC-ID',
-            'to': 'UniProtKB',
-            'ids': protein_queries
-        }
-        data = urllib.parse.urlencode(params)
-        data = data.encode('utf-8')
-        req = urllib.request.Request(url, data)
-
-        data_js = json.load(urllib_query(req))
-        job_id = data_js['jobId']
-
-        http_check_status = 'http://rest.uniprot.org/idmapping/status/{0}'.format(job_id)
-
-        check_status_response = requests.head(http_check_status)
-        check_status_response.raise_for_status()
-
-        # Not working because the code is 200 instead of 303...
-        # Issue with protein_queries formatting?
-        # Wait 5 secodns to check again return code.
-        # But is there no better way?
-        while check_status_response.status_code != 303:
-            time.sleep(5)
-            check_status_response = requests.head(http_check_status)
-
-        if check_status_response.status_code == 303:
-            http_download = 'http://rest.uniprot.org/idmapping/uniprotkb/results/stream/{0}'.format(job_id)
-            response = requests.get(http_download)
-            response.raise_for_status()
-            data = response.json()
-
-            results = {}
-            for result in data['results']:
-                protein_id = result['from']
-                protein_data = result['to']
-                reviewed = protein_data['entryType']
-                if 'reviewed' in reviewed:
-                    review = True
-                else:
-                    review = False
-                protein_description = protein_data['proteinDescription']
-                if 'recommendedName' in protein_description:
-                    if 'ecNumbers' in protein_description['recommendedName']:
-                        protein_ecs = list(set([ecnumber['value'] for ecnumber in protein_description['recommendedName']['ecNumbers']]))
-                    else:
-                        protein_ecs = []
-                    protein_fullname = protein_description['recommendedName']['fullName']['value']
+            protein_description = protein_data['proteinDescription']
+            if 'recommendedName' in protein_description:
+                if 'ecNumbers' in protein_description['recommendedName']:
+                    protein_ecs = list(set([ecnumber['value'] for ecnumber in protein_description['recommendedName']['ecNumbers']]))
                 else:
                     protein_ecs = []
-                    protein_fullname = ''
+                protein_fullname = protein_description['recommendedName']['fullName']['value']
+            else:
+                protein_ecs = []
+                protein_fullname = ''
 
-                gene_names = [gene['geneName']['value'] for gene in protein_data['genes'] if 'geneName' in gene]
-                if len(gene_names) > 0:
-                    gene_name = gene_names[0]
-                else:
-                    gene_name = ''
-                protein_xrefs = protein_data['uniProtKBCrossReferences']
+            gene_names = [gene['geneName']['value'] for gene in protein_data['genes'] if 'geneName' in gene]
+            if len(gene_names) > 0:
+                gene_name = gene_names[0]
+            else:
+                gene_name = ''
+            protein_xrefs = protein_data['uniProtKBCrossReferences']
 
-                rhea_ids = []
-                if 'comments' in protein_data:
-                    protein_comments = protein_data['comments']
-                    protein_reactions = [comment['reaction'] for comment in protein_comments if comment['commentType'] == 'CATALYTIC ACTIVITY']
-                    for reaction in protein_reactions:
-                        if 'reactionCrossReferences' in reaction:
-                            rhea_ids.extend([dbxref['id'] for dbxref in reaction['reactionCrossReferences'] if dbxref['database'] == 'Rhea' and 'RHEA:' in dbxref['id']])
-                        if 'ecNumber' in reaction:
-                            protein_ecs.append(reaction['ecNumber'])
-                    rhea_ids = list(set(rhea_ids))
-                protein_ecs = list(set(protein_ecs))
-                gos = list(set([xref['id'] for xref in protein_xrefs if xref['database'] == 'GO']))
-                interpros = list(set([xref['id'] for xref in protein_xrefs if xref['database'] == 'InterPro']))
-                results[protein_id] = [protein_fullname, review, gos, protein_ecs, interpros, rhea_ids, gene_name]
-                output_dict.update(results)
+            rhea_ids = []
+            if 'comments' in protein_data:
+                protein_comments = protein_data['comments']
+                protein_reactions = [comment['reaction'] for comment in protein_comments if comment['commentType'] == 'CATALYTIC ACTIVITY']
+                for reaction in protein_reactions:
+                    if 'reactionCrossReferences' in reaction:
+                        rhea_ids.extend([dbxref['id'] for dbxref in reaction['reactionCrossReferences'] if dbxref['database'] == 'Rhea' and 'RHEA:' in dbxref['id']])
+                    if 'ecNumber' in reaction:
+                        protein_ecs.append(reaction['ecNumber'])
+                rhea_ids = list(set(rhea_ids))
+            protein_ecs = list(set(protein_ecs))
+            gos = list(set([xref['id'] for xref in protein_xrefs if xref['database'] == 'GO']))
+            interpros = list(set([xref['id'] for xref in protein_xrefs if xref['database'] == 'InterPro']))
+            results[protein_id] = [protein_fullname, review, gos, protein_ecs, interpros, rhea_ids, gene_name]
+            output_dict.update(results)
 
     return output_dict
 
@@ -605,12 +700,11 @@ def search_already_annotated_protein(set_proteins, already_annotated_proteins, o
     return protein_to_search_on_uniprots, output_dict
 
 
-def query_uniprot_annotation_rest(protein_to_search_on_uniprots, beta, output_dict):
+def query_uniprot_annotation_rest(protein_to_search_on_uniprots, output_dict):
     """Query UniProt with REST to find protein annotations
 
     Args:
         protein_to_search_on_uniprots (set): set of proteins not already annotated that will need UniProt queries
-        beta (bool): option to use the new API of UniProt (in beta can be unstable)
         output_dict (dict): annotation dict: protein as key and annotation as value ([function_name, review_status, [go_terms], [ec_numbers], [interpros], [rhea_ids], gene_name])
 
     Returns:
@@ -619,21 +713,15 @@ def query_uniprot_annotation_rest(protein_to_search_on_uniprots, beta, output_di
     # The limit of 10 000 proteins per query comes from the help of Uniprot (inferior to 20 000):
     # https://www.uniprot.org/help/uploadlists
     if len(protein_to_search_on_uniprots) < 15000:
-        if not beta:
-            protein_queries = ' '.join(protein_to_search_on_uniprots)
-        else:
-            protein_queries = ','.join(protein_to_search_on_uniprots)
-        tmp_output_dict = rest_query_uniprot_to_retrieve_function(protein_queries, beta)
+        protein_queries = ','.join(protein_to_search_on_uniprots)
+        tmp_output_dict = rest_query_uniprot_to_retrieve_function(protein_queries)
         output_dict.update(tmp_output_dict)
         time.sleep(1)
     else:
         protein_chunks = chunks(list(protein_to_search_on_uniprots), 15000)
         for chunk in protein_chunks:
-            if not beta:
-                protein_queries = ' '.join(chunk)
-            else:
-                protein_queries = ','.join(chunk)
-            tmp_output_dict = rest_query_uniprot_to_retrieve_function(protein_queries, beta)
+            protein_queries = ','.join(chunk)
+            tmp_output_dict = rest_query_uniprot_to_retrieve_function(protein_queries)
             output_dict.update(tmp_output_dict)
             time.sleep(1)
 
@@ -791,7 +879,6 @@ def propagate_annotation_in_cluster(output_dict, reference_proteins, propagate_a
                 keep_gene_names = max(gene_names,key=gene_names.count)
             if len(protein_names) > 0:
                 keep_protein_names = max(protein_names,key=protein_names.count)
-
             protein_annotations[reference_protein] = [keep_protein_names, keep_gos, keep_ecs, keep_gene_names]
         else:
             # Use only annotation from representative proteins.
@@ -863,7 +950,7 @@ def write_pathologic_file(protein_annotations, pathologic_folder, base_filename,
         logger.critical('|EsMeCaTa|annotation| No reference proteins for {0}, esmecata will not create a pathologic folder for it.'.format(base_filename))
 
 
-def annotate_proteins(input_folder, output_folder, uniprot_sparql_endpoint, propagate_annotation, uniref_annotation, expression_annotation, beta=None):
+def annotate_proteins(input_folder, output_folder, uniprot_sparql_endpoint, propagate_annotation, uniref_annotation, expression_annotation):
     """Write the annotation associated with a cluster after propagation step into pathologic file for run on Pathway Tools.
 
     Args:
@@ -872,9 +959,9 @@ def annotate_proteins(input_folder, output_folder, uniprot_sparql_endpoint, prop
         propagate_annotation (float): float between 0 and 1. It is the ratio of proteins in the cluster that should have the annotation to keep this annotation.
         uniref_annotation (bool): option to use uniref annotation to add annotation
         expression_annotation (bool): option to add expression annotation from uniprot
-        beta (bool): option to use the new API of UniProt (in beta can be unstable)
     """
     starttime = time.time()
+    logger.info('|EsMeCaTa|annotation| Begin annotation.')
 
     if uniprot_sparql_endpoint is None and uniref_annotation is not None:
         logger.critical('|EsMeCaTa|annotation| At this moment, --uniref option needs to be used with --sparql option.')
@@ -936,7 +1023,10 @@ def annotate_proteins(input_folder, output_folder, uniprot_sparql_endpoint, prop
     # a second that use the annotation from the annotation files associated with the proteins.
     # It is slower as it reads the annotation file to retrieve the annotation, but I think it is less heavy on the memory.
     already_annotated_proteins_in_file = {}
-    for input_file in os.listdir(reference_protein_path):
+
+    input_files = os.listdir(reference_protein_path)
+    for index, input_file in enumerate(input_files):
+        logger.info('|EsMeCaTa|annotation| Annotation of {0} ({1} on {2} data to annotate).'.format(input_file, index+1, len(input_files)))
         base_file = os.path.basename(input_file)
         base_filename = os.path.splitext(base_file)[0]
 
@@ -948,12 +1038,11 @@ def annotate_proteins(input_folder, output_folder, uniprot_sparql_endpoint, prop
         #protein_to_search_on_uniprots, output_dict = search_already_annotated_protein(set_proteins, already_annotated_proteins, output_dict)
         # Second method to handle protein already annotated
         protein_to_search_on_uniprots, output_dict = search_already_annotated_protein_in_file(set_proteins, already_annotated_proteins_in_file, annotation_folder, output_dict)
-
         if uniprot_sparql_endpoint:
             proteomes = input_proteomes[base_filename]
             output_dict = query_uniprot_annotation_sparql(proteomes, base_filename, uniprot_sparql_endpoint, output_dict)
         else:
-            output_dict = query_uniprot_annotation_rest(protein_to_search_on_uniprots, beta, output_dict)
+            output_dict = query_uniprot_annotation_rest(protein_to_search_on_uniprots, output_dict)
 
         # First method to handle protein already annotated
         #already_annotated_proteins.update({protein: output_dict[protein] for protein in output_dict if protein not in already_annotated_proteins})
