@@ -20,12 +20,14 @@ import math
 import os
 import pandas as pd
 import random
+import re
 import requests
 import shutil
 import sys
 import time
 
 from collections import OrderedDict
+from requests.adapters import HTTPAdapter, Retry
 
 from Bio import __version__ as biopython_version
 from Bio import SeqIO
@@ -45,6 +47,41 @@ REQUESTS_HEADERS = {'User-Agent': 'EsMeCaTa proteomes v' + esmecata_version + ',
 logger = logging.getLogger(__name__)
 
 
+def get_next_link(headers):
+    """ From batch queries, get the next link to request.
+    Function from: https://www.uniprot.org/help/id_mapping
+
+    Args:
+        headers (dict): headers of batch response.
+
+    Returns:
+        str: next link to query in the batch process.
+    """
+    re_next_link = re.compile(r'<(.+)>; rel="next"')
+    if "Link" in headers:
+        match = re_next_link.match(headers["Link"])
+        if match:
+            return match.group(1)
+
+
+def get_batch(session, batch_url):
+    """ Batch queries.
+    Function from: https://www.uniprot.org/help/id_mapping
+
+    Args:
+        session (requests Session object): session used to query UniProt.
+        batch_url (str): URL of batch queries.
+
+    Returns:
+        batch_response (requests Response object): response to batch query.
+    """
+    while batch_url:
+        response = session.get(batch_url)
+        response.raise_for_status()
+        yield response
+        batch_url = get_next_link(response.headers)
+
+
 def taxonomic_affiliation_to_taxon_id(observation_name, taxonomic_affiliation, ncbi=None):
     """ From a taxonomic affiliation (such as cellular organisms;Bacteria;Proteobacteria;Gammaproteobacteria) find corresponding taxon ID for each taxon.
 
@@ -263,7 +300,7 @@ def requests_query(http_str, nb_retry=5):
         return response
 
 
-def rest_query_proteomes(observation_name, tax_id, tax_name, busco_percentage_keep, all_proteomes, beta=None):
+def rest_query_proteomes(observation_name, tax_id, tax_name, busco_percentage_keep, all_proteomes, session=None):
     """REST query on UniProt to get the proteomes associated with a taxon.
 
     Args:
@@ -272,12 +309,18 @@ def rest_query_proteomes(observation_name, tax_id, tax_name, busco_percentage_ke
         tax_name (str): taxon name associated with the tax_id
         busco_percentage_keep (float): BUSCO score to filter proteomes (proteomes selected will have a higher BUSCO score than this threshold)
         all_proteomes (bool): Option to select all the proteomes (and not only preferentially reference proteomes)
-        beta (bool): option to use the new API of UniProt (in beta can be unstable)
+        session: request session object
+
     Returns:
         proteomes (list): list of proteome IDs associated with the taxon ID
         organism_ids (dict): organism ID (key) associated with each proteomes (values)
         proteomes_data (list): list of lists with proteome_id, busco_score, assembly_level, org_tax_id, reference_proteome
     """
+    if not session:
+        retries = Retry(total=5, backoff_factor=0.25, status_forcelist=[500, 502, 503, 504])
+        session = requests.Session()
+        session.mount("https://", HTTPAdapter(max_retries=retries))
+
     proteomes = []
     proteomes_data = []
     organism_ids = {}
@@ -285,123 +328,64 @@ def rest_query_proteomes(observation_name, tax_id, tax_name, busco_percentage_ke
     # Take reference proteomes with "reference:yes".
     # Avoid redundant and excluded proteomes with "redundant%3Ano+excluded%3Ano".
     # Use "format=tab" to easily handle the ouput.
-    http_str = 'https://www.uniprot.org/proteomes/?query=taxonomy:{0}+reference:yes+redundant%3Ano+excluded%3Ano&format=tab'.format(tax_id)
-    beta_httpt_str = 'https://rest.uniprot.org/beta/proteomes/stream?query=(taxonomy_id%3A{0})AND(proteome_type%3A1)&format=json'.format(tax_id)
+    httpt_str = 'https://rest.uniprot.org/proteomes/stream?query=(taxonomy_id%3A{0})AND(proteome_type%3A1)&format=json&size=500'.format(tax_id)
 
     if all_proteomes:
-        http_str = 'https://www.uniprot.org/proteomes/?query=taxonomy:{0}+redundant%3Ano+excluded%3Ano&format=tab'.format(tax_id)
-        beta_httpt_str = 'https://rest.uniprot.org/beta/proteomes/stream?query=(taxonomy_id%3A{0})AND(proteome_type%3A2)&format=json'.format(tax_id)
+        httpt_str = 'https://rest.uniprot.org/proteomes/stream?query=(taxonomy_id%3A{0})AND(proteome_type%3A2)&format=json&size=500'.format(tax_id)
 
     # If esmecata does not find proteomes with only reference, search for all poroteomes even if they are not reference.
-    all_http_str = 'https://www.uniprot.org/proteomes/?query=taxonomy:{0}+redundant%3Ano+excluded%3Ano&format=tab'.format(tax_id)
-    all_beta_httpt_str = 'https://rest.uniprot.org/beta/proteomes/stream?query=(taxonomy_id%3A{0})AND(proteome_type%3A2)&format=json'.format(tax_id)
+    all_http_str = 'https://rest.uniprot.org/proteomes/stream?query=(taxonomy_id%3A{0})AND(proteome_type%3A2)&format=json&size=500'.format(tax_id)
 
-    response_proteome_status = False
+    data = {}
+    data['results'] = []
+    for batch_reponse in get_batch(session, httpt_str):
+        batch_json = batch_reponse.json()
+        data['results'].extend(batch_json['results'])
 
-    if not beta:
-        proteome_response = requests_query(http_str)
-        # Raise error if we have a bad request.
-        proteome_response.raise_for_status()
-        proteome_response_text = proteome_response.text
-        if proteome_response_text != '':
-            csvreader = csv.reader(proteome_response_text.splitlines(), delimiter='\t')
-            response_proteome_status = True
-            # Avoid header.
-            next(csvreader)
-        else:
-            csvreader = []
-        reference_proteome = True
+    reference_proteome = True
+    if len(data['results']) == 0:
+        logger.info('|EsMeCaTa|proteomes| {0}: No reference proteomes found for {1} ({2}) try non-reference proteomes.'.format(observation_name, tax_id, tax_name))
+        time.sleep(1)
+        for batch_reponse in get_batch(session, all_http_str):
+            batch_json = batch_reponse.json()
+            data['results'].extend(batch_json['results'])
+        reference_proteome = False
 
-        if response_proteome_status is False:
-            time.sleep(1)
-            logger.info('|EsMeCaTa|proteomes| {0}: No reference proteomes found for {1} ({2}) try non-reference proteomes.'.format(observation_name, tax_id, tax_name))
-            proteome_response = requests_query(all_http_str)
-            proteome_response_text = proteome_response.text
-            if proteome_response_text != '':
-                csvreader = csv.reader(proteome_response_text.splitlines(), delimiter='\t')
-                response_proteome_status = True
-                # Avoid header.
-                next(csvreader)
+    for proteome_data in data['results']:
+        proteome_id = proteome_data['id']
+        if 'proteomeCompletenessReport' in proteome_data:
+            if 'buscoReport' in proteome_data['proteomeCompletenessReport']:
+                proteome_busco = proteome_data['proteomeCompletenessReport']['buscoReport']
             else:
-                csvreader = []
-            reference_proteome = False
-
-        for line in csvreader:
-            proteome = line[0]
-            completness = line[6]
-            org_tax_id = line[2]
-
-            if line[4] != '':
-                busco_percentage = float(line[4].split(':')[1].split('%')[0])
-            else:
-                busco_percentage = None
-
-            # Check that proteome has busco score.
-            if busco_percentage_keep:
-                if busco_percentage and busco_percentage >= busco_percentage_keep and completness == 'full':
-                    proteomes.append(proteome)
-                    if org_tax_id not in organism_ids:
-                        organism_ids[org_tax_id] = [proteome]
-                    else:
-                        organism_ids[org_tax_id].append(proteome)
-            else:
-                if completness == 'full':
-                    proteomes.append(proteome)
-                    if org_tax_id not in organism_ids:
-                        organism_ids[org_tax_id] = [proteome]
-                    else:
-                        organism_ids[org_tax_id].append(proteome)
-
-            proteomes_data.append([proteome, busco_percentage, completness, org_tax_id, reference_proteome])
-    else:
-        proteome_response = requests_query(beta_httpt_str)
-        proteome_response.raise_for_status()
-        check_code = proteome_response.status_code
-        data = proteome_response.json()
-        reference_proteome = True
-        if len(data['results']) == 0:
-            logger.info('|EsMeCaTa|proteomes| {0}: No reference proteomes found for {1} ({2}) try non-reference proteomes.'.format(observation_name, tax_id, tax_name))
-            time.sleep(1)
-            proteome_response = requests_query(all_beta_httpt_str)
-            proteome_response.raise_for_status()
-            data = proteome_response.json()
-            reference_proteome = False
-
-        for proteome_data in data['results']:
-            proteome_id = proteome_data['id']
-            if 'proteomeCompletenessReport' in proteome_data:
-                if 'buscoReport' in proteome_data['proteomeCompletenessReport']:
-                    proteome_busco = proteome_data['proteomeCompletenessReport']['buscoReport']
-                else:
-                    proteome_busco = None
-                if proteome_busco:
-                    busco_score = (proteome_busco['complete'] / proteome_busco['total']) * 100
-                else:
-                    busco_score = None
+                proteome_busco = None
+            if proteome_busco:
+                busco_score = (proteome_busco['complete'] / proteome_busco['total']) * 100
             else:
                 busco_score = None
+        else:
+            busco_score = None
 
-            if 'level' in proteome_data['genomeAssembly']:
-                assembly_level = proteome_data['genomeAssembly']['level']
-            else:
-                assembly_level = ''
-            org_tax_id = proteome_data['taxonomy']['taxonId']
-            proteome_type = proteome_data['proteomeType']
-            if busco_percentage_keep:
-                if busco_score and busco_score >= busco_percentage_keep and assembly_level == 'full':
-                    proteomes.append(proteome_id)
-                    if org_tax_id not in organism_ids:
-                        organism_ids[org_tax_id] = [proteome_id]
-                    else:
-                        organism_ids[org_tax_id].append(proteome_id)
-            else:
-                if assembly_level == 'full':
-                    proteomes.append(proteome_id)
-                    if org_tax_id not in organism_ids:
-                        organism_ids[org_tax_id] = [proteome_id]
-                    else:
-                        organism_ids[org_tax_id].append(proteome_id)
-            proteomes_data.append([proteome_id, busco_score, assembly_level, org_tax_id, reference_proteome])
+        if 'level' in proteome_data['genomeAssembly']:
+            assembly_level = proteome_data['genomeAssembly']['level']
+        else:
+            assembly_level = ''
+        org_tax_id = proteome_data['taxonomy']['taxonId']
+        proteome_type = proteome_data['proteomeType']
+        if busco_percentage_keep:
+            if busco_score and busco_score >= busco_percentage_keep and assembly_level == 'full':
+                proteomes.append(proteome_id)
+                if org_tax_id not in organism_ids:
+                    organism_ids[org_tax_id] = [proteome_id]
+                else:
+                    organism_ids[org_tax_id].append(proteome_id)
+        else:
+            if assembly_level == 'full':
+                proteomes.append(proteome_id)
+                if org_tax_id not in organism_ids:
+                    organism_ids[org_tax_id] = [proteome_id]
+                else:
+                    organism_ids[org_tax_id].append(proteome_id)
+        proteomes_data.append([proteome_id, busco_score, assembly_level, org_tax_id, reference_proteome])
 
     return proteomes, organism_ids, proteomes_data
 
@@ -415,7 +399,6 @@ def sparql_query_proteomes(observation_name, tax_id, tax_name, busco_percentage_
         tax_name (str): taxon name associated with the tax_id
         busco_percentage_keep (float): BUSCO score to filter proteomes (proteomes selected will have a higher BUSCO score than this threshold)
         all_proteomes (bool): Option to select all the proteomes (and not only preferentially reference proteomes)
-        beta (bool): option to use the new API of UniProt (in beta can be unstable)
         uniprot_sparql_endpoint (str): uniprot SPARQL endpoint to query (by default query Uniprot SPARQL endpoint)
     Returns:
         proteomes (list): list of proteome IDs associated with the taxon ID
@@ -592,7 +575,7 @@ def subsampling_proteomes(organism_ids, limit_maximal_number_proteomes, ncbi):
 
 def find_proteomes_tax_ids(json_taxonomic_affiliations, ncbi, proteomes_description_folder,
                         busco_percentage_keep=None, all_proteomes=None, uniprot_sparql_endpoint=None,
-                        limit_maximal_number_proteomes=99, beta=None, minimal_number_proteomes=1):
+                        limit_maximal_number_proteomes=99, minimal_number_proteomes=1, session=None):
     """Find proteomes associated with taxonomic affiliations
 
     Args:
@@ -603,8 +586,8 @@ def find_proteomes_tax_ids(json_taxonomic_affiliations, ncbi, proteomes_descript
         all_proteomes (bool): Option to select all the proteomes (and not only preferentially reference proteomes)
         uniprot_sparql_endpoint (str): uniprot SPARQL endpoint to query (by default query Uniprot SPARQL endpoint)
         limit_maximal_number_proteomes (int): int threshold after which a subsampling will be performed on the data
-        beta (bool): option to use the new API of UniProt (in beta can be unstable)
         minimal_number_proteomes (int): minimal number of proteomes required to be associated with a taxon for the taoxn to be kepp
+        session: request session object
 
     Returns:
         proteomes_ids (dict): observation name (key) associated with proteome IDs
@@ -643,7 +626,7 @@ def find_proteomes_tax_ids(json_taxonomic_affiliations, ncbi, proteomes_descript
             if uniprot_sparql_endpoint:
                 proteomes, organism_ids, data_proteomes = sparql_query_proteomes(observation_name, tax_id, tax_name, busco_percentage_keep, all_proteomes, uniprot_sparql_endpoint)
             else:
-                proteomes, organism_ids, data_proteomes = rest_query_proteomes(observation_name, tax_id, tax_name, busco_percentage_keep, all_proteomes, beta)
+                proteomes, organism_ids, data_proteomes = rest_query_proteomes(observation_name, tax_id, tax_name, busco_percentage_keep, all_proteomes, session)
 
             for data_proteome in data_proteomes:
                 proteomes_descriptions.append([tax_id, tax_name, *data_proteome])
@@ -786,7 +769,7 @@ def compute_stat_proteomes(result_folder, stat_file=None):
 
 def retrieve_proteomes(input_file, output_folder, busco_percentage_keep=80,
                         ignore_taxadb_update=None, all_proteomes=None, uniprot_sparql_endpoint=None,
-                        remove_tmp=None, limit_maximal_number_proteomes=99, beta=None, rank_limit=None,
+                        remove_tmp=None, limit_maximal_number_proteomes=99, rank_limit=None,
                         minimal_number_proteomes=1):
     """From a tsv file with taxonomic affiliations find the associated proteomes.
 
@@ -799,12 +782,15 @@ def retrieve_proteomes(input_file, output_folder, busco_percentage_keep=80,
         all_proteomes (bool): Option to select all the proteomes (and not only preferentially reference proteomes)
         uniprot_sparql_endpoint (str): uniprot SPARQL endpoint to query (by default query Uniprot SPARQL endpoint)
         limit_maximal_number_proteomes (int): int threshold after which a subsampling will be performed on the data
-        beta (bool): option to use the new API of UniProt (in beta can be unstable)
         rank_limit (str): rank limit to remove from the data
         minimal_number_proteomes (int): minimal number of proteomes required to be associated with a taxon for the taoxn to be kepp
     """
     starttime = time.time()
 
+    retries = Retry(total=5, backoff_factor=0.25, status_forcelist=[500, 502, 503, 504])
+    session = requests.Session()
+    session.mount("https://", HTTPAdapter(max_retries=retries))
+
     if is_valid_file(input_file) is False:
         logger.critical('|EsMeCaTa|proteomes| The input {0} is not a valid file pathname.'.format(input_file))
         sys.exit()
@@ -868,7 +854,8 @@ def retrieve_proteomes(input_file, output_folder, busco_percentage_keep=80,
 
     if not os.path.exists(proteome_tax_id_file):
         proteomes_ids, single_proteomes, tax_id_not_founds = find_proteomes_tax_ids(json_taxonomic_affiliations, ncbi, proteomes_description_folder,
-                                                        busco_percentage_keep, all_proteomes, uniprot_sparql_endpoint, limit_maximal_number_proteomes, beta, minimal_number_proteomes)
+                                                        busco_percentage_keep, all_proteomes, uniprot_sparql_endpoint,
+                                                        limit_maximal_number_proteomes, minimal_number_proteomes, session)
 
         proteome_to_download = []
         for proteomes_id in proteomes_ids:
@@ -911,19 +898,17 @@ def retrieve_proteomes(input_file, output_folder, busco_percentage_keep=80,
     tmp_folder = os.path.join(output_folder, 'tmp_proteome')
     is_valid_dir(tmp_folder)
 
-    for proteome in proteome_to_download:
+    for index, proteome in enumerate(proteome_to_download):
+        logger.info('|EsMeCaTa|proteomes| Downloaded {0} on {1} proteomes'.format(index+1, len(proteome_to_download)))
         output_proteome_file = os.path.join(tmp_folder, proteome+'.faa.gz')
         if not os.path.exists(output_proteome_file):
             if uniprot_sparql_endpoint is not None:
                 sparql_get_protein_seq(proteome, output_proteome_file, uniprot_sparql_endpoint)
             else:
-                if beta:
-                    http_str = 'https://rest.uniprot.org/beta/uniprotkb/stream?query=proteome:{0}&format=fasta&compressed=true'.format(proteome)
-                else:
-                    http_str = 'https://www.uniprot.org/uniprot/?query=proteome:{0}&format=fasta&compress=yes'.format(proteome)
-                proteome_response = requests_query(http_str)
-                with open(output_proteome_file, 'wb') as f:
-                    f.write(proteome_response.content)
+                http_str = 'https://rest.uniprot.org/uniprotkb/stream?query=proteome:{0}&format=fasta&compressed=true&size=500'.format(proteome)
+                for batch_reponse in get_batch(session, http_str):
+                    with open(output_proteome_file, 'wb') as f:
+                        f.write(batch_reponse.content)
         time.sleep(1)
 
     # Download Uniprot metadata and create a json file containing them.
